use nom::branch::alt;
use nom::bytes::complete::is_not;
use nom::bytes::complete::tag_no_case;
use nom::character::complete::line_ending;
use nom::character::complete::space0;
use nom::character::complete::space1;
use nom::combinator::eof;
use nom::combinator::not;
use nom::combinator::opt;
use nom::combinator::verify;
use nom::multi::many_till;
use nom::sequence::tuple;

use super::org_source::OrgSource;
use super::util::in_section;
use crate::context::parser_with_context;
use crate::context::ContextElement;
use crate::context::ContextMatcher;
use crate::context::ExitClass;
use crate::context::ExitMatcherNode;
use crate::context::RefContext;
use crate::error::CustomError;
use crate::error::MyError;
use crate::error::Res;
use crate::parser::element_parser::element;
use crate::parser::util::blank_line;
use crate::parser::util::exit_matcher_parser;
use crate::parser::util::get_consumed;
use crate::parser::util::start_of_line;
use crate::types::Element;
use crate::types::GreaterBlock;
use crate::types::Paragraph;
use crate::types::SetSource;

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
pub fn greater_block<'b, 'g, 'r, 's>(
    context: RefContext<'b, 'g, 'r, 's>,
    input: OrgSource<'s>,
) -> Res<OrgSource<'s>, GreaterBlock<'s>> {
    // TODO: Do I need to differentiate between different greater block types.
    start_of_line(input)?;
    let (remaining, _leading_whitespace) = space0(input)?;
    let (remaining, (_begin, name)) = tuple((
        tag_no_case("#+begin_"),
        verify(name, |name: &OrgSource<'_>| {
            match Into::<&str>::into(name).to_lowercase().as_str() {
                "comment" | "example" | "export" | "src" | "verse" => false,
                _ => true,
            }
        }),
    ))(remaining)?;
    let context_name = match Into::<&str>::into(name).to_lowercase().as_str() {
        "center" => "center block".to_owned(),
        "quote" => "quote block".to_owned(),
        name @ _ => format!("special block {}", name),
    };
    if in_section(context, context_name.as_str()) {
        return Err(nom::Err::Error(CustomError::MyError(MyError(
            "Cannot nest objects of the same element".into(),
        ))));
    }
    let exit_with_name = greater_block_end(name.into());
    let (remaining, parameters) = opt(tuple((space1, parameters)))(remaining)?;
    let (remaining, _nl) = line_ending(remaining)?;
    let contexts = [
        ContextElement::ConsumeTrailingWhitespace(true),
        ContextElement::Context(context_name.as_str()),
        ContextElement::ExitMatcherNode(ExitMatcherNode {
            class: ExitClass::Alpha,
            exit_matcher: &exit_with_name,
        }),
    ];
    let parser_context = context.with_additional_node(&contexts[0]);
    let parser_context = parser_context.with_additional_node(&contexts[1]);
    let parser_context = parser_context.with_additional_node(&contexts[2]);
    let parameters = match parameters {
        Some((_ws, parameters)) => Some(parameters),
        None => None,
    };

    let element_matcher = parser_with_context!(element(true))(&parser_context);
    let exit_matcher = parser_with_context!(exit_matcher_parser)(&parser_context);
    // Check for a completely empty block
    let (remaining, children) = match tuple((
        not(exit_matcher),
        blank_line,
        many_till(blank_line, exit_matcher),
    ))(remaining)
    {
        Ok((remain, (_not_immediate_exit, first_line, (_trailing_whitespace, _exit_contents)))) => {
            let mut element = Element::Paragraph(Paragraph::of_text(first_line.into()));
            let source = get_consumed(remaining, remain);
            element.set_source(source.into());
            (remain, vec![element])
        }
        Err(_) => {
            let (remaining, (children, _exit_contents)) =
                many_till(element_matcher, exit_matcher)(remaining)?;
            (remaining, children)
        }
    };
    let (remaining, _end) = exit_with_name(&parser_context, remaining)?;

    // Not checking if parent exit matcher is causing exit because the greater_block_end matcher asserts we matched a full greater block

    let source = get_consumed(input, remaining);
    Ok((
        remaining,
        GreaterBlock {
            source: source.into(),
            name: name.into(),
            parameters: parameters.map(|val| Into::<&str>::into(val)),
            children,
        },
    ))
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
fn name<'s>(input: OrgSource<'s>) -> Res<OrgSource<'s>, OrgSource<'s>> {
    is_not(" \t\r\n")(input)
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
fn parameters<'s>(input: OrgSource<'s>) -> Res<OrgSource<'s>, OrgSource<'s>> {
    is_not("\r\n")(input)
}

fn greater_block_end<'c>(name: &'c str) -> impl ContextMatcher + 'c {
    // TODO: Can this be done without making an owned copy?
    move |context, input: OrgSource<'_>| _greater_block_end(context, input, name)
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
fn _greater_block_end<'b, 'g, 'r, 's, 'c>(
    _context: RefContext<'b, 'g, 'r, 's>,
    input: OrgSource<'s>,
    name: &'c str,
) -> Res<OrgSource<'s>, OrgSource<'s>> {
    start_of_line(input)?;
    let (remaining, _leading_whitespace) = space0(input)?;
    let (remaining, (_begin, _name, _ws)) = tuple((
        tag_no_case("#+end_"),
        tag_no_case(name),
        alt((eof, line_ending)),
    ))(remaining)?;
    let source = get_consumed(input, remaining);
    Ok((remaining, source))
}
